import numpy as np


def _calculate_coverage(self):
    spec_mat_lower_real = np.zeros_like(
        self.spectral_density_q[0], dtype=float
    )
    for j in range(self.n_freq):
        spec_mat_lower_real[j] = self._complex_to_real(
            self.spectral_density_q[0][j]
        )

    spec_mat_upper_real = np.zeros_like(
        self.spectral_density_q[2], dtype=float
    )
    for j in range(self.n_freq):
        spec_mat_upper_real[j] = self._complex_to_real(
            self.spectral_density_q[2][j]
        )
    coverage_point_CI = np.mean(
        (spec_mat_lower_real <= self.real_spec_true)
        & (self.real_spec_true <= spec_mat_upper_real)
    )
    return coverage_point_CI

    def _calculate_L2_error(self):
        spec_mat_median = self.spectral_density_q[1]
        N2_VI = np.empty(self.n_freq)
        for i in range(self.n_freq):
            N2_VI[i] = np.sum(
                np.diag(
                    (spec_mat_median[i, :, :] - self.spec_true[i, :, :])
                    @ (spec_mat_median[i, :, :] - self.spec_true[i, :, :])
                )
            )
        L2_VI = np.sqrt(np.mean(N2_VI))
        return L2_VI


def __get_uniform_ci(psd_all, psd_q):
    psd_median = psd_q[1]
    n_samples, n_freq, p, _ = psd_all.shape

    # transform elements of psd_all and psd_median to the real numbers
    real_psd_all = np.zeros_like(psd_all, dtype=float)
    real_psd_median = np.zeros_like(psd_median, dtype=float)

    for i in range(n_samples):
        for j in range(n_freq):
            real_psd_all[i, j] = __complex_to_real(psd_all[i, j])

    for j in range(n_freq):
        real_psd_median[j] = __complex_to_real(psd_median[j])

    # find the maximum normalized absolute deviation for real_psd_all
    max_std_abs_dev = __uniformmax_multi(real_psd_all)
    # find the threshold of the 90% as a screening criterion
    threshold = np.quantile(max_std_abs_dev, 0.9)

    # find the uniform CI for real_psd_median
    mad = median_abs_deviation(real_psd_all, axis=0, nan_policy="omit")
    mad[mad == 0] = 1e-10
    lower_bound = real_psd_median - threshold * mad
    upper_bound = real_psd_median + threshold * mad

    # Converts lower_bound and upper_bound to the complex matrix
    psd_uni_lower = np.zeros_like(lower_bound, dtype=complex)
    psd_uni_upper = np.zeros_like(upper_bound, dtype=complex)

    for i in range(n_freq):
        psd_uni_lower[i] = __real_to_complex(lower_bound[i])
        psd_uni_upper[i] = __real_to_complex(upper_bound[i])

    psd_uniform = np.stack([psd_uni_lower, psd_median, psd_uni_upper], axis=0)

    return psd_uniform


# Find the normalized median absolute deviation for every element among all smaples
# For all samples of each frequency, each matrix, return their maximum normalized absolute deviation
def __uniformmax_multi(mSample):
    N_sample, N, d, _ = mSample.shape
    C_help = np.zeros((N_sample, N, d, d))

    for j in range(N):
        for r in range(d):
            for s in range(d):
                C_help[:, j, r, s] = __uniformmax_help(mSample[:, j, r, s])

    return np.max(C_help, axis=0)


def __uniformmax_help(sample):
    return np.abs(sample - np.median(sample)) / median_abs_deviation(sample)


def __get_pointwise_ci(psd_all, quantiles):
    _, num_freq, p_dim, _ = psd_all.shape
    psd_q = np.zeros((3, num_freq, p_dim, p_dim), dtype=complex)

    diag_indices = np.diag_indices(p_dim)
    psd_q[:, :, diag_indices[0], diag_indices[1]] = np.quantile(
        np.real(psd_all[:, :, diag_indices[0], diag_indices[1]]),
        quantiles,
        axis=0,
    )

    # we dont do lower triangle because it is symmetric
    upper_triangle_idx = np.triu_indices(p_dim, k=1)
    real_part = np.real(
        psd_all[:, :, upper_triangle_idx[1], upper_triangle_idx[0]]
    )
    imag_part = np.imag(
        psd_all[:, :, upper_triangle_idx[1], upper_triangle_idx[0]]
    )

    for i, q in enumerate(quantiles):
        psd_q[i, :, upper_triangle_idx[1], upper_triangle_idx[0]] = (
            np.quantile(real_part, q, axis=0)
            + 1j * np.quantile(imag_part, q, axis=0)
        ).T

    psd_q[:, :, upper_triangle_idx[0], upper_triangle_idx[1]] = np.conj(
        psd_q[:, :, upper_triangle_idx[1], upper_triangle_idx[0]]
    )
    return psd_q



from starccato_jax.data import load_training_data

from starccato_sampler.sampler import sample


def test_sampler(outdir):
    _, val_data = load_training_data(train_fraction=0.8)
    sample(
        val_data[0],
        outdir=outdir,
        num_warmup=500,
        num_samples=1000,
        num_chains=1,
        verbose=True,
        stepping_stone_lnz=True,
    )



import matplotlib.pyplot as plt

from starccato_sampler.utils import beta_spaced_samples


def test_version():
    from starccato_sampler import __version__

    assert isinstance(__version__, str)


def test_beta_samples(outdir):
    plt.plot(beta_spaced_samples(100, 0.3, 1))
    plt.savefig(f"{outdir}/test_beta_samples.png")



import os

import arviz as az
import corner
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
from starccato_jax import StarccatoVAE
from starccato_jax.credible_intervals import coverage_probability, pointwise_ci
from starccato_jax.plotting import add_quantiles


def sampler_diagnostic_plots(inf_object: az.InferenceData, outdir: str):
    # Plot the trace plot
    axes = az.plot_trace(inf_object, var_names=["z"])
    if "true_latent" in inf_object.sample_stats:
        true_z = inf_object.sample_stats["true_latent"].to_numpy().ravel()

        if len(true_z) > 2:
            colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
            for i, _z in enumerate(true_z):
                axes[0, 0].axvline(
                    _z, color=colors[i % len(colors)], ls="--", alpha=0.3
                )
    axes[0, 0].set_ylim(bottom=0)
    plt.savefig(os.path.join(outdir, "trace_plot.png"))
    plt.close()

    # # Plot the corner plot
    # corner.corner(inf_object, var_names=["z"])
    # plt.savefig(os.path.join(outdir, "corner_plot.png"))
    # plt.close()


def plot_ci(
    y_obs: jnp.ndarray,
    z_posterior: jnp.ndarray,
    starccato_vae: StarccatoVAE,
    y_true: jnp.ndarray = None,
    nsamps=100,
    fname=None,
):
    z_samples = z_posterior[
        np.random.choice(z_posterior.shape[0], nsamps, replace=False)
    ]
    y_preds = starccato_vae.generate(z=z_samples)
    ypred_qtls = pointwise_ci(y_preds, ci=0.9)
    yrecn_qtls, yrecn_cov = None, None

    if y_true is not None:
        ypred_cov = coverage_probability(ypred_qtls, y_true)
        y_recon = starccato_vae.reconstruct(y_true, n_reps=nsamps)
        yrecn_qtls = pointwise_ci(y_recon, ci=0.9)
        yrecn_cov = coverage_probability(yrecn_qtls, y_true)

    plt.figure(figsize=(5, 3.5))
    plt.plot(
        y_obs, label="Observed", color="black", lw=2, zorder=-10, alpha=0.2
    )
    if y_true is not None:
        plt.plot(y_true, label="True", color="black", lw=2, zorder=0)
    ax = plt.gca()
    add_quantiles(
        ax, ypred_qtls, f"Posterior (coverage: {ypred_cov:.0%})", "tab:orange"
    )
    if yrecn_qtls is not None:
        add_quantiles(
            ax,
            yrecn_qtls,
            f"Reconstruction (coverage: {yrecn_cov:.0%})",
            "tab:green",
        )

    ax.set_xlim(0, len(y_obs))

    plt.legend(frameon=False)
    if fname is not None:
        plt.savefig(fname)



import dataclasses
import os
from typing import List

import jax.random as random
import matplotlib.pyplot as plt
import numpy as np
from lartillot_gaussian import LartillotGaussianModel

from starccato_sampler.evidence import (  # compute_harmonic_mean_evidence,
    compute_gaussian_approx_evidence,
    compute_stepping_stone_evidence,
)
from starccato_sampler.utils import beta_spaced_samples


@dataclasses.dataclass
class LnZEstimate:
    ln_z: float
    err: float
    label: str

    def __repr__(self):
        return f"{self.label}: {self.ln_z:.3f} +/- {self.err:.3f}"


def test_stepping_stone_evidence(outdir):
    d = 5
    lartillot_model = LartillotGaussianModel(d=d, v=1)
    rng = random.PRNGKey(0)
    ntemps = 32
    betas = beta_spaced_samples(ntemps, 0.3, 1)
    tempd_chains = lartillot_model.generate_chains(1000, betas)

    ss_lnz = LnZEstimate(
        *compute_stepping_stone_evidence(
            tempd_chains.lnl.T, betas, outdir, rng
        ),
        "Stepping Stone",
    )

    post_samples = lartillot_model.simulate_posterior_samples(5000, beta=1.0)
    gauss_lnz = LnZEstimate(
        *compute_gaussian_approx_evidence(
            ref_sample=np.zeros(d),
            lnl_ref=lartillot_model.log_likelihood(np.zeros((1, d))),
            posterior_samples=post_samples,
            n_bootstraps=30,
        ),
        "Gaussian Approx",
    )

    # chain = lartillot_model.generate_chains(1000, nchains=100)
    # harmonic_lnz = LnZEstimate(
    #     *compute_harmonic_mean_evidence(chains.samples, chains.lnl), "Harmonic Mean"
    # )

    lnzs = [ss_lnz, gauss_lnz]
    print(lnzs)
    plot_estimate(lartillot_model.lnZ, lnzs, outdir)

    plot_path = os.path.join(outdir, "lnz_comparison.png")

    for lnz in lnzs:
        np.testing.assert_allclose(
            lnz.ln_z,
            lartillot_model.lnZ,
            atol=0.3,
            err_msg=f"{lnz} not close to {lartillot_model.lnZ} within tolerance",
        )

    assert os.path.exists(plot_path), "Plot file should be created"


def plot_estimate(
    true_lnZ: float,
    lnzs: List[LnZEstimate],
    outdir: str = "out",
):
    plt.figure(figsize=(2 * len(lnzs), 3))

    for i, lnz in enumerate(lnzs):
        plt.errorbar(
            i,
            lnz.ln_z,
            yerr=lnz.err,
            fmt="o",
            color="tab:orange",
            label=lnz.label,
        )

    ylims = plt.ylim()

    # max_err = max([lnz.err for lnz in lnzs])
    #
    # # ensure ylims at least +- 2 from true_lnZ
    # if true_lnZ - max_err < ylims[0]:
    #     plt.ylim(true_lnZ - max_err, true_lnZ + max_err)
    #
    # plt.ylim(true_lnZ - max_err, true_lnZ + max_err)
    # set tick labels
    plt.xticks(range(len(lnzs)), [lnz.label for lnz in lnzs])
    plt.axhline(true_lnZ, color="r", label="True LnZ")
    plt.ylabel("LnZ")
    plt.tight_layout()
    plt.savefig(os.path.join(outdir, "lnz_comparison.png"))
